/***************************************************************************
 *   Copyright (C) 2009 by Stanson <stanson@pangolin.ru>                   *
 *   GPL v3 code, read COPYING file                                        *
 *                                                                         *
 * Audi Gamma III SAA1057 serial interface.                                *
 *                                                                         *
 **************************************************************************/

#include <pic18fregs.h>
#include "main.h"
#include "saa1057.h"

#define SAA1057_DELAY_uS(x) \
{ \
    TMR1H = HIGH(-(x) * 3);\
    TMR1L = LOW(-(x) * 3);\
    PIR1bits.TMR1IF = 0;\
    while( PIR1bits.TMR1IF == 0 );\
}

#define SAA1057_SET_DELAY_uS(x) \
{ \
    TMR1H = HIGH(-(x) * 3);\
    TMR1L = LOW(-(x) * 3);\
    PIR1bits.TMR1IF = 0;\
}

#define SAA1057_WAIT_DELAY { while( PIR1bits.TMR1IF == 0 ); }

/* F = Tuning_step * WordA * ( FM ? 10 : 1 )
    WordA = 512...32767
    Tuning_step = 1.25 or 1.00 kHz
    WordA(F) = F / ( FM ? 10 : 1 ) / Tuning_step
    WordA(95.2MHz) = 9520
*/

#define WORD_B			0x8100 /* word B with last 8 bits enabled ( SB2 = 1 ) */

/* FM/AM mode */
#define SAA1057_FM		0x4000
#define SAA1057_AM		0x0000

/* reference frequency selection ( 40 or 32 kHz ) */
#define SAA1057_STEP_125	0x2000	/* 40 kHz - tuning steps 1.25 kHz @ AM and 12.5 kHz @ FM */
#define SAA1057_STEP_100	0x0000	/* 32 kHz - tuning steps 1.00 kHz @ AM and 10.0 kHz @ FM */

/* programmable current amplifier gain */
#define SAA1057_GAIN_P1		0x0000	/* 0.023 */
#define SAA1057_GAIN_P2		0x0200	/* 0.07  */
#define SAA1057_GAIN_P3		0x0400	/* 0.23  */
#define SAA1057_GAIN_P4		0x0C00	/* 0.7   */
#define SAA1057_GAIN_P5		0x1C00	/* 2.3   */

/* latch A load mode */
#define SAA1057_SLA_SYNC	0x0080 /* syncronous mode for changes less than 31 step */

/* phase detector mode */
#define SAA1057_PDM_AUTO	0x0000
#define SAA1057_PDM_ON		0x0040
#define SAA1057_PDM_OFF		0x0060

/* bus receiver power save mode */
#define SAA1057_BRM		0x0010

/* TEST output mode */
#define SAA1057_TEST_1		0x0000	/* logical '1' */
#define SAA1057_TEST_REF	0x0004	/* reference frequency ( 32 or 40 kHz ) */
#define SAA1057_TEST_CNT	0x0001	/* programmable counter output */
#define SAA1057_TEST_LOCK	0x0005	/* in-lock output ( '0' = out-lock, '1' = in-lock ) */

word saa1057_B, saa1057_A;

void saa1057_write( word d )
{
    byte i;
    word r = d;

    if( d & 0x8000 ) saa1057_B = d;
    else             saa1057_A = d;

    /* wait for tDIST from previous load */
    SAA1057_WAIT_DELAY;

    SAA1057_CLB = 0;
    SAA1057_DATA = 0;
			/* leading zero */
    SAA1057_DELAY_uS(1);	/* tCLBlead = 1uS */
    SAA1057_DLEN = 1;
    SAA1057_DELAY_uS(5);	/* tCLBlag1 = 5uS */
    SAA1057_CLB = 1;
    SAA1057_DELAY_uS(5);	/* tCLBH = 5uS */
    SAA1057_CLB = 0;
    SAA1057_DELAY_uS(3);

    for( i = 0; i < 16; i++ )
    {
	if( r & 0x8000 ) SAA1057_DATA = 1;
	else             SAA1057_DATA = 0;

	SAA1057_DELAY_uS(2);
	SAA1057_CLB = 1;
	SAA1057_DELAY_uS(5);
	SAA1057_CLB = 0;
	SAA1057_DELAY_uS(3);
	r <<= 1;
    }

    /* load pulse */

    SAA1057_DATA = 0;
    SAA1057_DLEN = 0;
    SAA1057_DELAY_uS(2); /* tCLBlag2 = 2uS */
    SAA1057_CLB = 1;
    SAA1057_DELAY_uS(5); /* tCLBH = 5uS */
    SAA1057_CLB = 0;
    SAA1057_DELAY_uS(5);  /* minimum delay before next clock pulse: tDIST = 5uS */
    if( d & 0x8000 ) return; /* this was word B, don't have to wait */

    /* set up timer for SAA1057 busy time after word A transmission */
    if( saa1057_B & 0x0080 ) { SAA1057_SET_DELAY_uS(1300); }	/*  sync: tDIST = 1300uS */
    else                     { SAA1057_SET_DELAY_uS(300); }	/* async: tDIST = 300uS */
}

void saa1057_freq( word f )
{
    short diff;

    if( saa1057_B & SAA1057_FM ) f += 1070;
    else                         f += 450;

    if( f >= 512 && f <= 32767 )
    {
	diff = saa1057_A - f;
	if( diff < 0 ) diff = -diff;
	if( diff > 31 )
	{
	    if( saa1057_B & SAA1057_SLA_SYNC )
	    {
		saa1057_B &= ~SAA1057_SLA_SYNC;
		saa1057_write( saa1057_B );
	    }
	}
	saa1057_write( f & 0x7FFF );
    }
}

void saa1057_set( byte band, byte scan, word freq )
{
    word b = WORD_B |
	     SAA1057_PDM_AUTO |
	     SAA1057_BRM |
	     SAA1057_TEST_CNT |
	     SAA1057_GAIN_P2;

    if( band ) b |= SAA1057_FM;
    if( scan ) b |= SAA1057_SLA_SYNC;

    saa1057_write( b );
    saa1057_freq( freq );
}

void saa1057_init( void )
{
    TRISBbits.TRISB0 = 0;
    TRISBbits.TRISB1 = 0;
    TRISBbits.TRISB3 = 0;

    /* Set TMR1 (16 , CLK 3  (48  / 4 / 4) 0.333 ) */
    T1CON = 0xA1;

    SAA1057_DLEN = 0;

    saa1057_set( 1, 0, 0 );
}
